# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/markups.ipynb.

# %% auto 0
__all__ = ['Config', 'isnamedtupleinstance', 'unpack', 'AttrDict', 'pretty_json', 'read_json', 'write_json', 'write_jsonl',
           'read_jsonl', 'read_yaml', 'write_yaml', 'read_xml', 'write_xml']

# %% ../nbs/markups.ipynb 2
import json
import jsonlines
import yaml

from .loader import BB, L, np
from .paths import *
from .logger import *
import xmltodict
from typing import Union
from .thinc_parser.parser import Config

# %% ../nbs/markups.ipynb 3
def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def unpack(obj):
    if isinstance(obj, dict):
        return {key: unpack(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [unpack(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: unpack(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(unpack(value) for value in obj)
    else:
        return obj


class AttrDict(object):
    """Utility to call a dictionary like an object  
    FEATURES:  
    0. Get and Set any key (or a nested key) like it's an object, tab-completion supported  
    `self.key1.key2[0].key3` ...   
    1. Keys and Values are recursively parsed as AttrDicts  
    2. Formatting with self.pretty()  
    3. Convert everything to dict at any time via `self.to_dict()`
    3. Recursively drop keys using `self.drop(key)` from a json object  
    4. Recursively call a function on all values at all levels using `map`

    GOTCHAS:  
    1. Due to enforcing a `self.key` format all integer keys are converted to strings implicitly.   
    2. You can still call `self[int]` but this internally converts the int to string  
    """

    def __init__(self, data):
        for name, value in data.items():
            setattr(self, str(name), self._wrap(value))

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def _wrap(self, value):
        if isinstance(value, (L, tuple, list, set, frozenset)):
            value = type(value)([self._wrap(v) for v in value])
            if isinstance(value, (list, L)):
                value = L(value)
            return value
        else:
            return AttrDict(value) if isinstance(value, dict) else value

    __getitem__ = lambda self, x: getattr(self, str(x))
    __setitem__ = lambda self, k, v: setattr(self, str(k), self._wrap(v))

    def update(self, dict):
        for k,v in dict.items():
            self[k] = v

    def get(self, key, default=None):
        key = str(key)
        return self[key] if key in self else default

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return "{%s}" % str(
            ", ".join("'%s': %s" % (k, repr(v)) for (k, v) in self.__dict__.items())
        )

    def __dir__(self):
        return self.__dict__.keys()

    def __contains__(self, key):
        key = str(key)
        return key in self.__dict__.keys()

    def __delitem__(self, key):
        key = str(key)
        del self.__dict__[key]

    def map(self, func):
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.map(func)
            elif isinstance(v, (L, tuple, list, set, frozenset)):
                v = [func(_v) for _v in v if isinstance(_v, AttrDict)]
            else:
                v = func(v)
            self[k] = v

    def drop(self, key):
        if key in self:
            del self[key]
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.drop(key)
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.drop(key) for _v in v if isinstance(_v, AttrDict)]

    def to_dict(self):
        d = {}
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v = v.to_dict()
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.to_dict() if isinstance(_v, AttrDict) else _v for _v in v]
            d[k] = v
        return d

    def pretty(self, print_with_logger=False, *args, **kwargs):
        pretty_json(
            self.to_dict(), print_with_logger=print_with_logger, *args, **kwargs
        )

    def __eq__(self, other):
        return AttrDict(other).to_dict() == self.to_dict()


def pretty_json(i, fpath=None, indent=4, print_with_logger=True, return_as_string=False):
    def set_default(obj):
        if isinstance(obj, (set, BB, L)):
            return list(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, P):
            return str(obj)

    assert isinstance(i, (dict, list))
    i = unpack(i)
    dump = json.dumps(i, indent=indent, default=set_default)
    if fpath:
        makedir(parent(fpath))
        logger.info(f"Dumped a pretty file to {fpath}")
        with open(fpath, "w") as f:
            json.dump(i, f, indent=indent, default=set_default)
            return
    if print_with_logger:
        return logger.opt(depth=1).log("DEBUG", f"\n{dump}")
    else:
        if return_as_string:
            return dump
        print(dump)


# %% ../nbs/markups.ipynb 7
def read_json(fpath):
    import json

    with open(fpath, "r") as f:
        return json.load(f)


def write_json(obj, fpath, silent=False):
    def set_default(obj):
        if isinstance(obj, set):
            return list(obj)
        raise TypeError

    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Dumping json to {fpath}")
    with open(fpath, "w") as f:
        json.dump(obj, f, indent=4, default=set_default)
    return P(fpath)

# %% ../nbs/markups.ipynb 9
def write_jsonl(items, dest, mode='w'):
    makedir(parent(dest))
    with jsonlines.open(dest, mode) as writer:
        writer.write_all(items)
        if mode == 'a':
            Info(f"Appended {len(items)} items to {dest}")
        elif mode == 'w':
            Info(f"Wrote {len(items)} jsons to {dest}")


def read_jsonl(file):
    return [json.loads(line) for line in readlines(file, silent=True)]

# %% ../nbs/markups.ipynb 10
def read_yaml(file):
    with open(file, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def write_yaml(content, fpath):
    with open(fpath, "w") as outfile:
        yaml.dump(content, outfile, default_flow_style=False)

# %% ../nbs/markups.ipynb 11
def read_xml(file_path: Union[str, P]) -> AttrDict:
    "Read xml data as a dictionary"
    with open(str(file_path)) as xml_file:
        data = AttrDict(xmltodict.parse(xml_file.read()))
    return data


def write_xml(data: Union[AttrDict, dict], file_path: Union[str, P]):
    makedir(parent(file_path))
    "convert a dictionary to xml"
    with open(file_path, "w") as xml_file:
        data = data.to_dict() if isinstance(data, AttrDict) else data
        assert isinstance(data, dict), "Function only supports dicts for now"
        data = xmltodict.unparse(data, pretty=True)
        xml_file.write(data)

# %% ../nbs/markups.ipynb 12
Config = Config



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/misc.ipynb.

# %% auto 0
__all__ = ['Timer', 'timeit', 'io']

# %% ../nbs/misc.ipynb 2
import time
from .logger import Info
from .inspector import inspect

# %% ../nbs/misc.ipynb 3
class Timer:
    def __init__(self, N):
        "print elapsed time every iteration and print out remaining time"
        "assumes this timer is called exactly N times or less"
        self.start = time.time()
        self.N = N
        self.ix = 0

    def __call__(self, ix=None, info=None):
        ix = self.ix if ix is None else ix
        info = "" if info is None else f"{info}\t"
        elapsed = time.time() - self.start
        speed = elapsed / (ix + 1)
        unit = 's/iter'
        if speed < 1:
            speed = 1 / speed
            unit = 'iters/s'
        print(
            "{}{}/{} ({:.2f}s - {:.2f}s remaining - {:.2f} {}){}".format(
                info, ix + 1, self.N, elapsed, (self.N - ix) * (elapsed / (ix + 1)), speed, unit, " "*10
            ),
            end="\r",
        )
        self.ix += 1

# %% ../nbs/misc.ipynb 5
def timeit(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Info(f"{time.time() - s:.2f} seconds to execute `{func.__name__}`")
        return o

    return inner


def io(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Info(f"Args: {inspect(args)}\nKWargs: {inspect(kwargs)}\nOutput: {inspect(o)}")
        return o

    return inner




__version__ = "0.499.33"
from .loader import *
from .paths import *
from .markup import *
from .inspector import *
from .load_defaults import *
from .pdf_loader import PDF
from .ipython import *
from .decorators import *

try:
    from .torch_loader import *
except Exception as e:
    ...



__all__ = [
    "textify",
    "find_lines",
    "find_blocks",
    "find_substring",
    "get_line_data_from_word_data",
    "edit_distance_path",
    "group_blocks",
]
import string

import pandas as pd, numpy as np
from fuzzywuzzy import fuzz
from sklearn.cluster import dbscan
from statistics import mean
import matplotlib.pyplot as plt

from torch_snippets import *


def textify(dataframe: pd.DataFrame, im=None, separator_size: int = 20):
    """Create text out of a dataframe returned by im2df"""
    assert isinstance(dataframe, pd.DataFrame)
    if dataframe is None:
        return "", 0
    show = True if im is not None else False
    dataframe = left_to_right_top_to_bottom(dataframe)
    bbs = df2bbs(dataframe)
    if show:
        fig, ax = plt.subplots(1, figsize=(20, 20))
        ax.imshow(im, cmap="gray")
    TEXT = ""
    prev_top = 0
    prev_left = 0
    for ix, row in dataframe.iterrows():
        x, y, X, Y = bb = bbs[ix]
        l, t = x, y
        w, h = bb.w, bb.h
        curr_top = t
        curr_left = l
        sep = (
            "\n"
            if (curr_top - prev_top > separator_size) or (curr_left - prev_left > 500)
            else " "
        )
        text = row["text"]
        text = "" if text is None else text
        TEXT = TEXT + sep + text
        prev_top = t
        prev_left = l
        if show:
            ax.text(l, t - 20, "<{}>".format(row["text"]))
            # TODO: What is patches?
            rect = None
            # rect = patches.Rectangle((l, t), w, h, angle=0.0, linewidth=np.random.randint(2, 5),
            #                          edgecolor=np.random.rand(3), facecolor='none')
            ax.add_patch(rect)
    if show:
        plt.show()
    return TEXT


def find_lines(df=None, eps=20):
    df = df.reset_index()
    if "line" in df.columns:
        df.drop("line", inplace=True, axis=1)
    bbs = df2bbs(df)
    extra_columns = df.columns.tolist()
    for col in ["bb", "x", "y", "X", "Y"]:
        if col in extra_columns:
            extra_columns.remove(col)
    extra_columns = {col: df[col] for col in extra_columns}

    df = find_lines_for_bbs(bbs, eps=eps)
    for col in extra_columns:
        df = pd.merge(df, extra_columns[col], left_on="i", right_index=True)
    df.drop(["i"], inplace=True, axis=1)
    df.columns = [
        *"xyXY",
        "line",
        *extra_columns.keys(),
    ]  # "x,y,X,Y,line,text,conf,type".split(",")
    return df


def find_blocks(df, eps=20):
    blocks = []
    for line in sorted(df["line"].unique()):
        _df = df[df["line"] == line].sort_values("x")
        if line % 1 != 0:
            _df["line_block"] = line
        else:
            _df["delta_x"] = _df["x"] - _df["X"].shift(1)
            _df["line_block"] = (
                df["line"].map(str)
                + "_"
                + ((_df["x"] - _df["X"].shift(1)) > eps).cumsum().map(str)
            )
        blocks.append(_df)
    return pd.concat(blocks)


def find_substring(needle, hay):
    hay = hay.replace("\n", " ")
    needle_length = len(needle.split())
    max_sim_val = 0
    max_sim_string = ""
    for ix, ngram in enumerate(ngrams(hay.split(), needle_length)):
        hay_ngram = " ".join(ngram)
        similarity = SM(None, hay_ngram, needle).ratio()
        if similarity > max_sim_val:
            max_sim_val = similarity
            max_sim_string = hay_ngram

    for ix, ngram in enumerate(
        ngrams(hay.split(), needle_length + int(0.25 * needle_length))
    ):
        hay_ngram = " ".join(ngram)
        similarity = SM(None, hay_ngram, needle).ratio()
        if similarity > max_sim_val:
            max_sim_val = similarity
            max_sim_string = hay_ngram

    for ix, ngram in enumerate(
        ngrams(hay.split(), needle_length + int(0.5 * needle_length))
    ):
        hay_ngram = " ".join(ngram)
        similarity = SM(None, hay_ngram, needle).ratio()
        if similarity > max_sim_val:
            max_sim_val = similarity
            max_sim_string = hay_ngram

    for ix, ngram in enumerate(
        ngrams(hay.split(), needle_length + int(0.75 * needle_length))
    ):
        hay_ngram = " ".join(ngram)
        similarity = SM(None, hay_ngram, needle).ratio()
        if similarity > max_sim_val:
            max_sim_val = similarity
            max_sim_string = hay_ngram

    if max_sim_string.strip() == "":
        return "", 0, [hay]
    rest_of_strings = hay.split(max_sim_string)
    rest_of_strings = [s.strip() for s in rest_of_strings if s.strip() != ""]
    return max_sim_string, max_sim_val, rest_of_strings


def find_lines_for_bbs(bbs, eps=20):
    """Find Lines For Identified Bounding Boxes"""
    if len(bbs) == 0:
        return pd.DataFrame(columns=list("xyXY"))
    df = pd.DataFrame([(ix, x, y, X, Y) for ix, (x, y, X, Y) in enumerate(bbs)])
    df.columns = list("ixyXY")
    ys = []
    for bb in bbs:
        ys.append((bb[1], bb[3]))
    _ys = np.array(ys)
    db = dbscan(_ys, eps, min_samples=2)
    lines = np.c_[_ys, db[1]]
    # Applied dbscan to the y,Y values
    df["line_"] = db[1]
    df_ = df.groupby("line_").agg({"Y": "mean"})
    df_ = df_.sort_values("Y", ascending=True)
    for index in df_.index:
        if index == -1:
            df_.drop([-1], axis=0, inplace=True)
    # Created a dataframe that contains mean Y value for each row
    line__ = {}
    lctr = 0
    for line_ in df_.index:
        line__[line_] = lctr
        lctr += 1
    line__[-1] = -1
    # Created a dictionary to check if the numbering goes wrong
    df["line"] = df.line_.map(lambda x: line__[x])
    df_.index = df_.index.map(lambda x: line__[x])
    # Mapped the row numbers with the dictionary numbers
    for row in range(len(df)):
        if df.loc[row, "line_"] != -1:
            continue
        tmp = df_[df_["Y"] < df.loc[row, "Y"]]
        if tmp.index.max() is np.nan:
            df.loc[row, "line"] = -0.5
        else:
            df.loc[row, "line"] = tmp.index.max() + 0.5
    df = df.drop(columns=["line_"])
    df.sort_values(["line", "X"], inplace=True)
    df = df.reset_index(drop=True)
    df1 = df[df["line"] % 1 == 0.5].reset_index()
    df2 = df[df["line"] % 1 != 0.5]
    idx = list(df1["index"])
    df1 = df1.drop(columns=["index"]).sort_values(["Y"])
    df1["index"] = idx
    df2 = df2.reset_index()
    df = pd.concat([df2, df1], axis=0, sort=False)
    df.sort_values(["index"], inplace=True)
    df = df.drop(columns=["index"]).reset_index(drop=True)
    better_lines(df)
    return df


def better_lines(df):
    for line in df[df.line % 1 == 0.5].line.unique():
        _x = df[df["line"] == line]
        s, e = line - 0.5, line + 0.5
        new_lines = np.linspace(s, e, len(_x) + 2)[1:-1]
        df.loc[_x.index, "line"] = new_lines


def divide_text_lines(df, eps=10):
    if len(df) < 1:
        return df
    elif len(df) == 1:
        df["line"] = 0
    else:
        line = [None] * len(df)
        line_count = 0
        for r_idx, row in df.iterrows():
            if r_idx == 0:
                text, conf, x, y, X, Y = row
                line[r_idx] = line_count
            else:
                text1, conf1, x1, y1, X1, Y1 = row
                if abs(y - y1) < eps:
                    line[r_idx] = line_count
                else:
                    line_count += 1
                    line[r_idx] = line_count
                    text, conf, x, y, X, Y = row
        df["line"] = line
    df = df.sort_values(["line", "x"]).reset_index(drop=True)
    return df


def left_to_right_top_to_bottom(df):
    if "line" in [c.lower() for c in df.columns]:
        return df.sort_values(["line", "x"])
    else:
        return df.sort_values(["y", "x"])


def im2phrases(df, im=None, seperator_size: int = 20):
    if len(df) == 0:
        return "", 0
    show = False if im is None else True
    if show:
        fig, ax = plt.subplots(1, figsize=(20, 20))
        ax.imshow(im, cmap="gray")
    TEXT, text = [], []
    POINTS, points = [], []
    CONF, conf = [], []
    for r_idx, row in df.iterrows():
        x, y, X, Y = row["x"], row["y"], row["X"], row["Y"]
        w, h = X - x, Y - y
        if r_idx == 0:
            prev_l, prev_t, prev_r, prev_b, prev_line, prev_text, prev_conf = (
                x,
                y,
                X,
                Y,
                row["line"],
                asciify(row["text"]),
                row["conf"],
            )
            points.append((prev_l, prev_t, prev_r, prev_b))
            text.append(prev_text)
            conf.append(prev_conf)
            continue
        else:
            curr_l, curr_t, curr_r, curr_b, curr_line, curr_text, curr_conf = (
                x,
                y,
                X,
                Y,
                row["line"],
                asciify(row["text"]),
                row["conf"],
            )

        if (curr_line != prev_line) or abs(prev_r - curr_l) > 20:
            TEXT.append(" ".join(text))
            POINTS.append(points)
            CONF.append(mean(conf))
            text, points, conf = [], [], []
            text.append(curr_text)
            points.append((curr_l, curr_t, curr_r, curr_b))
            conf.append(curr_conf)
        else:
            text.append(curr_text)
            points.append((curr_l, curr_t, curr_r, curr_b))
            conf.append(curr_conf)

        prev_l, prev_t, prev_r, prev_b, prev_line = (
            curr_l,
            curr_t,
            curr_r,
            curr_b,
            curr_line,
        )
    if r_idx == len(df) - 1:
        TEXT.append(" ".join(text))
        POINTS.append(points)
        CONF.append(mean(conf))
    df = pd.DataFrame(
        data=list(zip(TEXT, POINTS, CONF)),
        columns=["text_lines", "line_ends(x,y,X,Y)", "conf"],
    )

    return df


def asciify(string):
    return "".join([c for c in string if c in valid_chars])


def find_match(word, words, N=1, lower=False):
    if word is None:
        return None, None
    if lower:
        word = word.lower()
        words = [w.lower() for w in words]
    matches, lengths = [], []
    for ix, f in enumerate(words):
        # if len(word)<=len(f)//2 or len(f)<=len(word)//2:
        # 	matches.append(-1)
        # 	lengths.append(-1)
        # 	continue
        k = min(len(f), len(word)) + 2
        m = fuzz.ratio(word[:k], f[:k])
        matches.append(m)
        lengths.append(k)
    sorted_idx = np.argsort(matches)
    if N > 1:
        return sorted_idx[:N], [words[ix] for ix in sorted_idx[:N]]
    else:
        M = max(matches)
        ixs = [ix for ix, m in enumerate(matches) if m == M]
        if len(ixs) == 1:
            return np.argmax(matches), max(matches)
        else:
            ix = max(ixs, key=lambda x: lengths[x])
            ix = min(ixs)
            return ix, matches[ix]


from difflib import SequenceMatcher as SM
from nltk.util import ngrams


def fuzzyfind(needle, hay):
    needle_length = len(needle.split())
    max_sim_val = 0
    max_sim_string = ""

    for ngram in ngrams(hay.split(), needle_length + int(0.2 * needle_length)):
        hay_ngram = " ".join(ngram)

        # TODO: No reference to similarity
        # if similarity > max_sim_val:
        #     max_sim_val = similarity
        #     max_sim_string = hay_ngram

    return max_sim_val, max_sim_string, hay.find(max_sim_string)


valid_chars = string.ascii_letters + "1234567890 !@#$%^&*(){}[];':\",./<>?-_=+"


def get_line_data_from_word_data(df, line_gap=10, block_gap=10):
    df = find_lines(df, eps=line_gap)
    df = find_blocks(df, eps=block_gap)
    df = df.groupby("line_block").agg(
        {
            "x": min,
            "y": min,
            "X": max,
            "Y": max,
            "text": lambda x: " ".join(x),
            **{c: lambda x: list(x) for c in df.columns if c not in [*"xyXY", "text"]},
        }
    )
    return df


def edit_distance_path(string1, string2):
    if string1 is None or string2 is None:
        return -1, []
    m = len(string1)
    n = len(string2)

    # Create a matrix to store the edit distances
    distance_matrix = [[0] * (n + 1) for _ in range(m + 1)]

    # Initialize the first row and column
    for i in range(m + 1):
        distance_matrix[i][0] = i
    for j in range(n + 1):
        distance_matrix[0][j] = j

    # Calculate the edit distances
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if string1[i - 1] == string2[j - 1]:
                distance_matrix[i][j] = distance_matrix[i - 1][j - 1]
            else:
                distance_matrix[i][j] = min(
                    distance_matrix[i - 1][j] + 1,  # deletion
                    distance_matrix[i][j - 1] + 1,  # insertion
                    distance_matrix[i - 1][j - 1] + 1,  # substitution
                )

    # Trace back the edit distance path
    i = m
    j = n
    path = []

    while i > 0 and j > 0:
        if string1[i - 1] == string2[j - 1]:
            path.append((string1[i - 1], string2[j - 1], ""))
            i -= 1
            j -= 1
        else:
            if distance_matrix[i][j] == distance_matrix[i - 1][j] + 1:
                path.append((string1[i - 1], "", "delete"))
                i -= 1
            elif distance_matrix[i][j] == distance_matrix[i][j - 1] + 1:
                path.append(("", string2[j - 1], "insert"))
                j -= 1
            else:
                path.append((string1[i - 1], string2[j - 1], "replace"))
                i -= 1
                j -= 1

    while i > 0:
        path.append((string1[i - 1], "", "delete"))
        i -= 1

    while j > 0:
        path.append(("", string2[j - 1], "insert"))
        j -= 1

    # Reverse the path to get the correct order
    path.reverse()

    return distance_matrix[m][n], path


def group_blocks(df):
    output = []
    for block, _df in df.groupby("line_block"):
        texts = [t for t in _df.text if isinstance(t, str)]
        _block = {
            "text": " ".join(texts),
            "x": _df.x.min(),
            "y": _df.y.min(),
            "X": _df.X.max(),
            "Y": _df.Y.max(),
            "line_block": block,
        }
        output.append(_block)
    output = pd.DataFrame(output)
    return output
